@@ -1,6 +1,7 @@
 {-# OPTIONS -fffi #-}
 module External (
     copyFileOrUrl, copyFilesOrUrls,
+    cloneTree, cloneTreeExcept, clonePartialsTree, clonePaths,
     fetchFilePS, gzFetchFilePS,
     sendEmail, sendEmailDoc, resendEmail,
     signString, verifyPS,
@@ -12,12 +13,15 @@ module External (
   ) where
 
 import List ( intersperse )
-import Monad ( liftM, when )
+import Monad ( liftM, when, zipWithM_ )
 import System ( ExitCode(..), system, getEnv )
 import IO ( hPutStr, hPutStrLn, hClose, hFlush, try, stdout )
+import System.IO.Error ( isDoesNotExistError )
 import System.IO.Unsafe ( unsafePerformIO )
+import System.Posix ( getSymbolicLinkStatus, isRegularFile, isDirectory )
+import System.Directory ( createDirectory, getDirectoryContents )
 import Char ( toUpper )
-import Foreign.C ( CString, withCString )
+import Foreign.C ( CString, withCString, CInt )
 import Foreign.Ptr ( nullPtr )
 #ifdef HAVE_CURSES
 import Foreign.C ( CChar )
@@ -25,7 +29,7 @@ import Foreign.Ptr ( Ptr )
 import Foreign.Marshal.Alloc (allocaBytes)
 import Autoconf ( use_color )
 #endif
-import Workaround ( createLink )
+import Workaround ( createLink, createDirectoryIfMissing )
 
 import DarcsFlags ( DarcsFlag( Quiet, SignAs, Sign, SignSSL,
                                Verbose, Verify, VerifySSL ) )
@@ -69,8 +73,69 @@ copyFileOrUrl fou _   _     = fail $ "unknown transport protocol: " ++ fou
 
 copyLocal  :: String -> FilePath -> IO ()
 copyLocal fou out =
-    createLink fou out `catchall` do c <- readFilePS fou
-                                     writeFilePS out c
+    createLink fou out `catchall` cloneFile fou out
+
+clonePaths :: FilePath -> FilePath -> [FilePath] -> IO ()
+clonePaths source dest = mapM_ (clonePath source dest)
+
+clonePath :: FilePath -> FilePath -> FilePath -> IO ()
+clonePath source dest path
+ = do let source' = source ++ "/" ++ path
+          dest' = dest ++ "/" ++ path
+      fs <- getSymbolicLinkStatus source'
+      if isDirectory fs then do
+          createDirectoryIfMissing True dest'
+       else if isRegularFile fs then do
+          createDirectoryIfMissing True (dest ++ "/" ++ basename path)
+          cloneFile source' dest'
+       else fail ("clonePath: Bad file " ++ source')
+   `catch` fail ("clonePath: Bad file " ++ source ++ "/" ++ path)
+ where basename = reverse . dropWhile ('/' /=) . dropWhile ('/' ==) . reverse
+
+clonePartialsTree :: FilePath -> FilePath -> [FilePath] -> IO ()
+clonePartialsTree source dest = mapM_ (clonePartialTree source dest)
+
+clonePartialTree :: FilePath -> FilePath -> FilePath -> IO ()
+clonePartialTree source dest "" = cloneTree source dest
+clonePartialTree source dest pref
+ = do createDirectoryIfMissing True (dest ++ "/" ++ basename pref)
+      cloneSubTree (source ++ "/" ++ pref) (dest ++ "/" ++ pref)
+ where basename = reverse . dropWhile ('/' /=) . dropWhile ('/' ==) . reverse
+
+cloneTree :: FilePath -> FilePath -> IO ()
+cloneTree = cloneTreeExcept []
+
+cloneTreeExcept :: [FilePath] -> FilePath -> FilePath -> IO ()
+cloneTreeExcept except source dest =
+ do fs <- getSymbolicLinkStatus source
+    if isDirectory fs then do
+        fps <- getDirectoryContents source
+        let fps' = filter (`notElem` (".":"..":except)) fps
+            mk_source fp = source ++ "/" ++ fp
+            mk_dest   fp = dest   ++ "/" ++ fp
+        zipWithM_ cloneSubTree (map mk_source fps') (map mk_dest fps')
+     else fail ("cloneTreeExcept: Bad source " ++ source)
+   `catch` fail ("cloneTreeExcept: Bad source " ++ source)
+
+cloneSubTree :: FilePath -> FilePath -> IO ()
+cloneSubTree source dest =
+ do fs <- getSymbolicLinkStatus source
+    if isDirectory fs then do
+        createDirectory dest
+        fps <- getDirectoryContents source
+        let fps' = filter (`notElem` [".", ".."]) fps
+            mk_source fp = source ++ "/" ++ fp
+            mk_dest   fp = dest   ++ "/" ++ fp
+        zipWithM_ cloneSubTree (map mk_source fps') (map mk_dest fps')
+     else if isRegularFile fs then do
+        cloneFile source dest
+     else fail ("cloneSubTree: Bad source "++ source)
+    `catch` (\e -> if isDoesNotExistError e
+                   then return ()
+                   else ioError e)
+
+cloneFile :: FilePath -> FilePath -> IO ()
+cloneFile source dest = readFilePS source >>= writeFilePS dest
 
 maybeURLCmd :: String -> String -> IO(Maybe(String))
 maybeURLCmd what url =
@@ -186,19 +251,23 @@ wgetRemotes u ns d = do wget_command <- getEnv "DARCS_WGET" `catch`
                                               "source files:"] ++ ns
 
 copySSHs :: String -> [String] -> FilePath -> IO ()
-copySSHs u ns d = do sftp_command <- getEnv "DARCS_SFTP" `catch`
+copySSHs u ns d = do p <- try $ getEnv "SSH_PORT" -- or DARCS_SSH_PORT ?
+                     sftp_command <- getEnv "DARCS_SFTP" `catch`
                                      \_ -> return "sftp"
-                     let sftp = head $ words sftp_command
-                         sftp_args = tail $ words sftp_command
+                     let port = either (const []) (\x->["-oPort="++x]) p
+                         sftp = head $ words sftp_command
+                         sftp_args = port ++ tail (words sftp_command)
                          path = drop 1 $ dropWhile (/= ':') u
                          host = (takeWhile (/= ':') u)++":"
                          cd = "cd "++path++"\n"
                          input = cd++(unlines $ map ("get "++) ns)
                      withCurrentDirectory d $ withOpenTemp $ \(th,tn) ->
+                         withTemp $ \sftpoutput ->
                          do hPutStr th input
                             hClose th
                             r <- exec sftp (sftp_args++["-b",tn,host])
-                                 "/dev/null" "/dev/null"
+                                 "/dev/null" sftpoutput
+                            output <- readFile sftpoutput
                             let files = if length ns > 5
                                           then (take 5 ns) ++ ["and "
                                                ++ (show (length ns - 5)) ++ " more"]
@@ -210,7 +279,8 @@ copySSHs u ns d = do sftp_command <- getEnv "DARCS_SFTP" `catch`
                                  fail $ unlines $
                                           ["(sftp) failed to fetch files.",
                                            "source directory: " ++ path,
-                                           "source files:"] ++ files ++ hint
+                                           "source files:"] ++ files ++
+                                          ["sftp output:",output] ++ hint
 
 
 copyRemoteCmd :: String -> FilePath -> IO ()
@@ -259,7 +329,7 @@ get_ext_cmd' = try_cmd cmds
                                    "--header=Cache-Control: no-cache",
                                    "-O",t,s])),
                   ("curl", (("--version",2),
-                            \t s -> ["curl", "-s", "-L",
+                            \t s -> ["curl", "-s", "-f", "-L",
                                      "-H", "Pragma: no-cache",
                                      "-H", "Cache-Control: no-cache",
                                      "-o",t,s]))]
@@ -537,12 +607,12 @@ can_execute (arg,expected_return_value) exe = do
   - or -1 if colour output not supported or error occured.
   - Terminal type determined by TERM env. variable.
   -}
-getTermNColors :: IO Int
+getTermNColors :: IO CInt
 
 #ifdef HAVE_CURSES
 
-foreign import ccall "term.h tgetnum" c_tgetnum :: CString -> IO Int
-foreign import ccall "term.h tgetent" c_tgetent :: Ptr CChar -> CString -> IO Int
+foreign import ccall "term.h tgetnum" c_tgetnum :: CString -> IO CInt
+foreign import ccall "term.h tgetent" c_tgetent :: Ptr CChar -> CString -> IO CInt
 
 termioBufSize :: Int
 termioBufSize = 4096
@@ -553,7 +623,7 @@ getTermNColors = if not use_color
                          allocaBytes termioBufSize (getTermNColorsImpl term)
                       `catch` \_ -> return (-1)
 
-getTermNColorsImpl :: String -> Ptr CChar -> IO Int
+getTermNColorsImpl :: String -> Ptr CChar -> IO CInt
 getTermNColorsImpl term buf = do rc <- withCString term $
                                        \termp -> c_tgetent buf termp
                                  if (rc /= 1) then return (-1)  else withCString "Co" $ \capap -> c_tgetnum capap
