@@ -1,8 +1,9 @@
-{-# OPTIONS -fffi #-}
+{-# OPTIONS -fffi -cpp #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  FastPackedString
--- Copyright   :  (c) The University of Glasgow 2001, David Roundy 2003
+-- Copyright   :  (c) The University of Glasgow 2001,
+--                    David Roundy 2003-2005
 -- License : GPL (I'm happy to also license this file BSD style but don't
 --           want to bother distributing two license files with darcs.
 --
@@ -19,82 +20,96 @@
 -- rewritten to support slices and use ForeignPtr by David Roundy
 
 module FastPackedString (
-	-- * The @PackedString@ type
-        PackedString,      -- abstract, instances: Eq, Ord, Show, Typeable
+        -- * The @PackedString@ type
+        PackedString,    -- abstract, instances: Eq, Ord, Show, Typeable
         unsafeWithInternals, -- :: PackedString -> (Ptr Word8 -> Int -> IO a) -> IO a
+        constructPS,
 
          -- * Converting to and from @PackedString@s
     generatePS,  -- :: Int -> (Ptr Word8 -> Int -> IO Int) -> IO PackedString
-	packString,  -- :: String -> PackedString
-	packWords,   -- :: [Word8] -> PackedString
-	unpackPS,    -- :: PackedString -> String
+        packString,  -- :: String -> PackedString
+        mallocedCString2PS, -- :: CString -> IO PackedString
+        withCStringPS, -- :: PackedString -> (CString -> IO a) -> IO a
+        packWords,   -- :: [Word8] -> PackedString
+        unpackPS,    -- :: PackedString -> String
         unpackWords, -- :: PackedString -> [Word8]
         unpackPSfromUTF8, -- :: PackedString -> String
 
-	-- * I\/O with @PackedString@s
-	hPutPS,      -- :: Handle -> PackedString -> IO ()
-	hGetPS,      -- :: Handle -> Int -> IO PackedString
+        -- * I\/O with @PackedString@s
+        hPutPS,      -- :: Handle -> PackedString -> IO ()
+        hGetPS,      -- :: Handle -> Int -> IO PackedString
     hGetContentsPS, -- :: Handle -> IO PackedString
-	readFilePS,  -- :: FilePath -> IO PackedString
-	writeFilePS, -- :: FilePath -> PackedString -> IO ()
-	gzReadFilePS,-- :: FilePath -> IO PackedString
-	mmapFilePS,  -- :: FilePath -> IO PackedString
+        readFilePS,  -- :: FilePath -> IO PackedString
+        writeFilePS, -- :: FilePath -> PackedString -> IO ()
+        gzReadFilePS,-- :: FilePath -> IO PackedString
+        gzReadFileLazily,  -- :: FilePath -> IO LazyFile
+        LazyFile(..),
+        mmapFilePS,  -- :: FilePath -> IO PackedString
+        gzWriteFilePS,  -- :: FilePath -> PackedString -> IO ()
+        gzWriteFilePSs, -- :: FilePath -> [PackedString] -> IO ()
 
-	-- * List-like manipulation functions
-	nilPS,       -- :: PackedString
-	consPS,      -- :: Char -> PackedString -> PackedString
-	headPS,      -- :: PackedString -> Char
-	tailPS,      -- :: PackedString -> PackedString
-	lastPS,      -- :: PackedString -> Char
-	nullPS,      -- :: PackedString -> Bool
-	appendPS,    -- :: PackedString -> PackedString -> PackedString
-	lengthPS,    -- :: PackedString -> Int
-	indexPS,     -- :: PackedString -> Int -> Char
-	indexPSW,    -- :: PackedString -> Int -> Word8
-	mapPS,       -- :: (Char -> Char) -> PackedString -> PackedString
-	--filterPS,    -- :: (Char -> Bool) -> PackedString -> PackedString
-	reversePS,   -- :: PackedString -> PackedString
-	concatPS,    -- :: [PackedString] -> PackedString
-	concatLenPS, -- :: Int -> [PackedString] -> PackedString
-	elemPS,      -- :: Char -> PackedString -> Bool
-	takePS,      -- :: Int -> PackedString -> PackedString
-	dropPS,      -- :: Int -> PackedString -> PackedString
-	splitAtPS,   -- :: Int -> PackedString -> (PackedString, PackedString)
+        -- * List-like manipulation functions
+        nilPS,       -- :: PackedString
+        consPS,      -- :: Char -> PackedString -> PackedString
+        headPS,      -- :: PackedString -> Char
+        tailPS,      -- :: PackedString -> PackedString
+        initPS,      -- :: PackedString -> PackedString
+        lastPS,      -- :: PackedString -> Char
+        nullPS,      -- :: PackedString -> Bool
+        appendPS,    -- :: PackedString -> PackedString -> PackedString
+        lengthPS,    -- :: PackedString -> Int
+        indexPS,     -- :: PackedString -> Int -> Char
+        indexPSW,    -- :: PackedString -> Int -> Word8
+        mapPS,       -- :: (Char -> Char) -> PackedString -> PackedString
+        --filterPS,    -- :: (Char -> Bool) -> PackedString -> PackedString
+        reversePS,   -- :: PackedString -> PackedString
+        concatPS,    -- :: [PackedString] -> PackedString
+        concatLenPS, -- :: Int -> [PackedString] -> PackedString
+        elemPS,      -- :: Char -> PackedString -> Bool
+        takePS,      -- :: Int -> PackedString -> PackedString
+        dropPS,      -- :: Int -> PackedString -> PackedString
+        splitAtPS,   -- :: Int -> PackedString -> (PackedString, PackedString)
 
-	foldlPS,     -- :: (a -> Char -> a) -> a -> PackedString -> a
-	foldrPS,     -- :: (Char -> a -> a) -> a -> PackedString -> a
+        foldlPS,     -- :: (a -> Char -> a) -> a -> PackedString -> a
+        foldrPS,     -- :: (Char -> a -> a) -> a -> PackedString -> a
         anyPS,
-	takeWhilePS, -- :: (Char -> Bool) -> PackedString -> PackedString
-	dropWhilePS, -- :: (Char -> Bool) -> PackedString -> PackedString
-	dropWhitePS, -- :: PackedString -> PackedString
-	breakWhitePS,-- :: PackedString -> Maybe (PackedString,PackedString)
-	spanPS,      -- :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
-	breakPS,     -- :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
-	breakOnPS,   -- :: Char -> PackedString -> (PackedString, PackedString)
-	linesPS,     -- :: PackedString -> [PackedString]
-	unlinesPS,     -- :: [PackedString] -> PackedString
+        takeWhilePS, -- :: (Char -> Bool) -> PackedString -> PackedString
+        dropWhilePS, -- :: (Char -> Bool) -> PackedString -> PackedString
+        dropWhitePS, -- :: PackedString -> PackedString
+        breakWhitePS,-- :: PackedString -> Maybe (PackedString,PackedString)
+        spanPS,      -- :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
+        spanEndPS,   -- :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
+        breakPS,     -- :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
+        breakOnPS,   -- :: Char -> PackedString -> (PackedString, PackedString)
+        linesPS,     -- :: PackedString -> [PackedString]
+        unlinesPS,     -- :: [PackedString] -> PackedString
         findPS,
+        hashPS,
 
-	wordsPS,     -- :: PackedString -> [PackedString]
-	splitPS,     -- :: Char -> PackedString -> [PackedString]
-	splitWithPS, -- :: (Char -> Bool) -> PackedString -> [PackedString]
+        wordsPS,     -- :: PackedString -> [PackedString]
+        splitPS,     -- :: Char -> PackedString -> [PackedString]
+        splitWithPS, -- :: (Char -> Bool) -> PackedString -> [PackedString]
 
---	joinPS,      -- :: PackedString -> [PackedString] -> PackedString
+--      joinPS,      -- :: PackedString -> [PackedString] -> PackedString
 
         breakFirstPS,-- :: Char -> PackedString -> Maybe (PackedString,PackedString)
         breakLastPS, -- :: Char -> PackedString -> Maybe (PackedString,PackedString)
+        breakFirstPairPS,
+        substrPS,    -- :: PackedString -> PackedString -> Maybe Int
         readIntPS,   -- :: PackedString -> Maybe (Int, PackedString)
-        mylexPS,     -- :: PackedString -> Maybe (PackedString, PackedString)
         is_funky,    -- :: PackedString -> Bool
         fromHex2PS,  -- :: PackedString -> PackedString
         fromPS2Hex,  -- :: PackedString -> PackedString
         betweenLinesPS,--  :: PackedString -> PackedString -> PackedString -> Maybe (PackedString)
+        break_after_nth_newline,
+        break_before_nth_newline,
     ) where
 
 import IO ( Handle, hClose, hFileSize, IOMode(ReadMode,WriteMode),
             hSeek, SeekMode(SeekFromEnd), hGetChar )
+import Control.Exception( bracket )
 
-import Autoconf ( use_mmap, use_hideous_malloc_hack )
+import Autoconf ( use_mmap )
 
 import System.Mem ( performGC )
 import Foreign.Storable ( peekElemOff, peek, poke )
@@ -104,24 +119,30 @@ import Foreign.Marshal.Array ( pokeArray, mallocArray, reallocArray,
                                peekArray, advancePtr )
 import Foreign.Marshal.Utils ( with )
 import Foreign.C.String
-import Foreign.C.Types ( CLong, CInt, CSize )
+import Foreign.C.Types ( CLong, CInt, CSize, )
+import Data.Bits ( rotateL )
 import Data.Char
 import Data.Word
+import Int ( Int32 )
 import Monad ( liftM, when )
-import Workaround ( handleToFd, fdToInt )
 
-import System.IO.Unsafe ( unsafePerformIO )
+import System.IO.Unsafe ( unsafePerformIO, unsafeInterleaveIO )
 import System.IO ( hPutBuf, hGetBuf )
 
-import System.IO ( openBinaryFile )
+import System.IO ( openBinaryFile, )
 
-import Foreign.ForeignPtr( ForeignPtr, withForeignPtr, mallocForeignPtrArray )
-import Foreign.Concurrent( newForeignPtr )
+import Foreign.Ptr ( castPtr )
+import Foreign.ForeignPtr( ForeignPtr, withForeignPtr, mallocForeignPtrArray,
+                           newForeignPtr )
+#if defined(__GLASGOW_HASKELL__)
+import qualified Foreign.Concurrent as FC ( newForeignPtr )
+import System.Posix ( handleToFd )
+#endif
 
 #ifdef DEBUG_PS
 import Foreign.ForeignPtr ( addForeignPtrFinalizer )
-import Foreign.Ptr ( FunPtr )
 #endif
+import Foreign.Ptr ( FunPtr )
 
 debugForeignPtr :: ForeignPtr a -> String -> IO ()
 #ifdef DEBUG_PS
@@ -137,19 +158,12 @@ debugForeignPtr fp n =
 debugForeignPtr _ _ = return ()
 #endif
 
-foreign import ccall unsafe "static fpstring.h dumb_malloc" dumb_malloc
-    :: Int -> IO (Ptr Word8)
-foreign import ccall unsafe "static stdio.h free" c_free
-    :: Ptr Word8 -> IO ()
+foreign import ccall unsafe "static stdio.h &free" c_free
+    :: FunPtr (Ptr Word8 -> IO ())
 mallocForeignPtr :: Int -> IO (ForeignPtr Word8)
 mallocForeignPtr l
     = do when (l > 1000000) performGC
-         if use_hideous_malloc_hack
-          then do p <- dumb_malloc $ max 1 l
-                  when (p == nullPtr) $
-                       fail $ "Memory allocation error! "++show l
-                  newForeignPtr p (c_free p)
-          else mallocForeignPtrArray l
+         mallocForeignPtrArray l
 
 ----------------------------------------------------------------------------
 --A way of creating ForeignPtrs outside the IO monad (althogh it still
@@ -178,6 +192,11 @@ unsafeWithInternals :: PackedString -> (Ptr Word8 -> Int -> IO a) -> IO a
 unsafeWithInternals (PS fp s l) f
  = withForeignPtr fp $ \p -> f (p `plusPtr` s) l
 
+constructPS
+  :: (Ptr Word8) -> Int -> IO () -> IO PackedString
+constructPS p l f = do fp <- FC.newForeignPtr p f
+                       return $ PS fp 0 l
+
 {-# INLINE (!) #-}
 (!) :: PackedString -> Int -> Word8
 (PS x s _l) ! i
@@ -198,48 +217,32 @@ generatePS i f
  = do p <- mallocArray i
       i' <- f p
       p' <- reallocArray p i'
-      fp <- newForeignPtr p' (c_free p')
+      fp <- newForeignPtr c_free p'
       return $ PS fp 0 i'
 
 instance Eq PackedString where
-   (==) = psniceq
+   (==) = eqPS
 
 foreign import ccall unsafe "static string.h memcmp" c_memcmp
     :: Ptr Word8 -> Ptr Word8 -> Int -> IO Int
 
-{-# INLINE psniceq #-}
-psniceq :: PackedString -> PackedString -> Bool
-psniceq a b | nullPS a && nullPS b = True
-psniceq (PS x1 s1 l1) (PS x2 s2 l2) =
-    (l1 == l2 &&) $ unsafePerformIO $ withForeignPtr x1 $ \p1->
-    withForeignPtr x2 $ \p2 ->
-        if p1 == p2 && s1 == s2
-        then return True
-        else liftM (==0) $ c_memcmp (p1 `plusPtr` s1) (p2 `plusPtr` s2) l1
+{-# INLINE eqPS #-}
+eqPS :: PackedString -> PackedString -> Bool
+eqPS a b = (comparePS a b) == EQ
 
 instance Ord PackedString where
-    compare = pscmp
+    compare = comparePS
 
-pscmp :: PackedString -> PackedString -> Ordering
-pscmp (PS x1 s1 l1) (PS x2 s2 l2) =
-    unsafePerformIO $ withForeignPtr x1 $ \p1->
-        withForeignPtr x2 $ \p2 ->
-    let doc :: Ptr Word8 -> Ptr Word8 -> IO Ordering
-        st1 = p1 `plusPtr` s1 `plusPtr` l1
-        st2 = p2 `plusPtr` s2 `plusPtr` l2
-        doc w1 w2 =
-            if w1 == st1 && w2 == st2 then return EQ
-            else if w1 == st1 then return LT
-                 else if w2 == st2 then return GT
-                      else do h1 <- peek w1
-                              h2 <- peek w2
-                              if h1 < h2
-                                 then return LT
-                                 else if h1 == h2
-                                      then doc (w1 `plusPtr` 1) (w2 `plusPtr` 1)
-                                      else return GT
-        in
-        doc (p1 `plusPtr` s1) (p2 `plusPtr` s2)
+-- | 'comparePS' provides an 'Ordering' for 'PackedStrings' supporting slices.
+comparePS :: PackedString -> PackedString -> Ordering
+comparePS (PS _ _ 0) (PS _ _ 0) = EQ    -- short cut for empty strings
+comparePS (PS x1 s1 l1) (PS x2 s2 l2) = unsafePerformIO $ 
+    withForeignPtr x1 $ \p1 -> 
+        withForeignPtr x2 $ \p2 -> do 
+            i <- c_memcmp (p1 `plusPtr` s1) (p2 `plusPtr` s2) (min l1 l2)
+            return $ case i `compare` 0 of
+                EQ  -> l1 `compare` l2
+                x   -> x
 
 --instance Read PackedString: ToDo
 
@@ -270,6 +273,28 @@ w2c = chr . fromIntegral
 c2w :: Char -> Word8
 c2w = fromIntegral . ord
 
+foreign import ccall unsafe "static string.h strlen" c_strlen
+    :: CString -> IO CInt
+foreign import ccall unsafe "static stdlib.h malloc" c_malloc
+    :: CInt -> IO (Ptr Word8)
+foreign import ccall unsafe "static stdlib.h free" free_cstring
+    :: CString -> IO ()
+
+
+mallocedCString2PS :: CString -> IO PackedString
+mallocedCString2PS cs = do fp <- newForeignPtr c_free (castPtr cs)
+                           l <- c_strlen cs
+                           return $ PS fp 0 (fromIntegral l)
+
+withCStringPS :: PackedString -> (CString -> IO a) -> IO a
+withCStringPS (PS ps s l) = bracket alloc free_cstring
+    where alloc = withForeignPtr ps $ \p ->
+                  do buf <- c_malloc (fromIntegral l+1)
+                     c_memcpy (castPtr buf) (castPtr p `plusPtr` s)
+                                  (fromIntegral l)
+                     poke (buf `plusPtr` l) (0::Word8)
+                     return $ castPtr buf
+
 -- -----------------------------------------------------------------------------
 -- Destructor functions (taking PackedStrings apart)
 
@@ -352,14 +377,22 @@ unsafeHeadPS :: PackedString -> Char
 unsafeHeadPS (PS x s _) -- ps ! 0 is inlined manually to eliminate a (+0)
   = w2c $ unsafePerformIO $ withForeignPtr x $ \p -> peekElemOff p s
 
+-- | Extract the elements after the head of a packed string, which must be non-empty.
 {-# INLINE tailPS #-}
 tailPS :: PackedString -> PackedString
-tailPS ps
-  | len <= 0 = error "FastPackedString.tailPS: tail []"
-  | len == 1 = nilPS
-  | otherwise  = substrPS ps 1 (len - 1)
-  where
-    len = lengthPS ps
+tailPS (PS p s l) 
+    | l <= 0    = error ("FastPackedString.tailPS: empty list")
+    | l == 1    = nilPS                                                                    
+    | otherwise = PS p (s+1) (l-1)
+
+-- | Return all the elements of a packed string except the last one.
+-- The string must be finite and non-empty.
+{-# INLINE initPS #-}
+initPS :: PackedString -> PackedString
+initPS (PS p s l) 
+    | l <= 0    = error ("FastPackedString.initPS: empty list")
+    | l == 1    = nilPS                                                                    
+    | otherwise = PS p s (l-1)                                                          
 
 {-# INLINE unsafeTailPS #-}
 unsafeTailPS :: PackedString -> PackedString
@@ -401,7 +434,7 @@ takePS n ps@(PS x s _) = if n >= lengthPS ps then ps
                          else PS x s n -- substrPS ps 0 (n - 1)
 
 {-# INLINE dropPS #-}
-dropPS	:: Int -> PackedString -> PackedString
+dropPS  :: Int -> PackedString -> PackedString
 dropPS n ps@(PS x s l)
     | n >= lengthPS ps = nilPS
     | otherwise = PS x (s+n) (l-n) -- substrPS ps n (lengthPS ps - 1)
@@ -428,6 +461,12 @@ findWhenPS f ps = seq f $
     else if f $ unsafeHeadPS ps then 0
          else 1 + findWhenPS f (unsafeTailPS ps)
 
+findFromEndUntilPS :: (Char -> Bool) -> PackedString -> Int
+findFromEndUntilPS f ps@(PS x s l) = seq f $
+    if nullPS ps then 0
+    else if f $ lastPS ps then l
+         else findFromEndUntilPS f (PS x s (l-1))
+
 {-# INLINE takeWhilePS #-}
 takeWhilePS :: (Char -> Bool) -> PackedString -> PackedString
 takeWhilePS f ps = seq f $ takePS (findWhenPS (not . f) ps) ps
@@ -464,6 +503,9 @@ elemPS c ps = c `elem` unpackPS ps
 spanPS :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
 spanPS  p ps = breakPS (not . p) ps
 
+spanEndPS :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
+spanEndPS  p ps = splitAtPS (findFromEndUntilPS (not.p) ps) ps
+
 breakPS :: (Char -> Bool) -> PackedString -> (PackedString, PackedString)
 breakPS p ps = case findWhenPS p ps of
                n -> (takePS n ps, dropPS n ps)
@@ -474,6 +516,31 @@ breakOnPS c p = case findPS c p of
                 Nothing -> (p,nilPS)
                 Just n -> (takePS n p, dropPS n p)
 
+{-# INLINE hashPS #-}
+hashPS :: PackedString -> Int32
+hashPS (PS x s l) =
+    unsafePerformIO $ withForeignPtr x $ \p->
+    do hash (p `plusPtr` s) l
+
+hash :: Ptr Word8 -> Int -> IO Int32
+hash ptr len = f (0 :: Int32) ptr len
+ where f h _ 0 = return h
+       f h p n = do x <- peek p
+                    let h' =  (fromIntegral x) + (rotateL h 8)
+                    f h' (p `advancePtr` 1) (n-1)
+
+{-# INLINE substrPS #-}
+substrPS :: PackedString -> PackedString -> Maybe Int
+substrPS tok str
+    | nullPS tok = Just 0
+    | lengthPS tok > lengthPS str = Nothing
+    | otherwise = do n <- findPS (headPS tok) str
+                     let ttok = tailPS tok
+                         reststr = dropPS (n+1) str
+                     if ttok == takePS (lengthPS ttok) reststr
+                        then Just n
+                        else ((n+1)+) `fmap` substrPS tok reststr
+
 {-# INLINE breakWhitePS #-}
 breakWhitePS :: PackedString -> (PackedString,PackedString)
 breakWhitePS (PS x s l) =
@@ -496,6 +563,16 @@ breakLastPS c p = case findLastPS c p of
                   Nothing -> Nothing
                   Just n -> Just (takePS n p, dropPS (n+1) p)
 
+breakFirstPairPS ::
+    Char -> Char -> PackedString -> Maybe (PackedString, PackedString)
+breakFirstPairPS c1 c2 ps =
+    let w1 = c2w c1
+        w2 = c2w c2
+    in case [ m | m <- [0..(lengthPS ps) - 2],
+                  (ps ! m) == w1, (ps ! (m + 1)) == w2 ] of
+       [] -> Nothing
+       (n:_) -> Just (takePS (n+1) ps, dropPS (n+1) ps)
+
 {-# INLINE linesPS #-}
 linesPS :: PackedString -> [PackedString]
 linesPS ps = case wfindPS (c2w '\n') ps of
@@ -526,7 +603,7 @@ concatPS xs
       p <- mallocArray start_size
       f p 0 1024 xs
     where f ptr len _ [] = do ptr' <- reallocArray ptr len
-                              fp <- newForeignPtr ptr' (c_free ptr')
+                              fp <- newForeignPtr c_free ptr'
                               return $ PS fp 0 len
           f ptr len to_go pss@(PS p s l:pss')
            | l <= to_go = do withForeignPtr p $ \pf ->
@@ -607,6 +684,7 @@ splitWithPS f ps =
 -- -----------------------------------------------------------------------------
 -- Local utility functions
 
+{-
 -- The definition of @_substrPS@ is essentially:
 -- @take (end - begin + 1) (drop begin str)@.
 
@@ -615,6 +693,7 @@ substrPS (PS ps s _) begin end = PS ps (s+begin) (1+end-begin)
 --substrPS (PS ps s l) begin end
 --    | end <= l && begin <= end && begin >= 0 = PS ps (s+begin) (1+end-begin)
 --    | otherwise = bug "substrPS out of bounds"
+-}
 
 -- -----------------------------------------------------------------------------
 -- hPutPS
@@ -660,7 +739,7 @@ hGetContentsPS h
       i <- hGetBuf h p start_size
       if i < start_size
        then do p' <- reallocArray p i
-               fp <- newForeignPtr p' (c_free p')
+               fp <- newForeignPtr c_free p'
                return $ PS fp 0 i
        else f p start_size
     where f p s = do let s' = 2 * s
@@ -668,7 +747,7 @@ hGetContentsPS h
                      i <- hGetBuf h (p' `plusPtr` s) s
                      if i < s then do let i' = s + i
                                       p'' <- reallocArray p' i'
-                                      fp <- newForeignPtr p'' (c_free p'')
+                                      fp <- newForeignPtr c_free p''
                                       return $ PS fp 0 i'
                               else f p' s'
 
@@ -712,6 +791,8 @@ foreign import ccall unsafe "static zlib.h gzclose" c_gzclose
     :: Ptr () -> IO ()
 foreign import ccall unsafe "static zlib.h gzread" c_gzread
     :: Ptr () -> Ptr Word8 -> Int -> IO Int
+foreign import ccall unsafe "static zlib.h gzwrite" c_gzwrite
+    :: Ptr () -> Ptr Word8 -> Int -> IO Int
 
 gzReadFilePS :: FilePath -> IO PackedString
 gzReadFilePS f = do
@@ -733,6 +814,53 @@ gzReadFilePS f = do
                  when (lread /= len) $ fail $ "problem gzreading file "++f
                  return $ PS fp 0 len
 
+data LazyFile = LazyString String
+              | MMappedPackedString PackedString
+              | LazyPackedStrings [PackedString]
+
+gzReadFileLazily :: FilePath -> IO LazyFile
+gzReadFileLazily f = do
+    h <- openBinaryFile f ReadMode
+    header <- hGetPS h 2
+    if header == packString "\31\139" then
+        do hClose h
+           withCString f $ \fstr-> withCString "rb" $ \rb-> do
+               gzf <- c_gzopen fstr rb
+               when (gzf == nullPtr) $ fail $ "problem opening file "++f
+               let read_rest = do
+                       -- We might be making too big a fp here
+                       fp <- mallocForeignPtr blocksize
+                       debugForeignPtr fp $ "gzReadFileLazily "++f
+                       lread <- withForeignPtr fp
+                              $ \p -> c_gzread gzf p blocksize
+                       case lread of
+                           0 -> do c_gzclose gzf
+                                   return []
+                           -1 -> fail $ "problem gzreading file "++f
+                           l | l < blocksize -> do c_gzclose gzf
+                                                   return [PS fp 0 l]
+                           l -> do rest <- unsafeInterleaveIO read_rest
+                                   return (PS fp 0 l:rest)
+               liftM LazyPackedStrings read_rest
+#if defined(__GLASGOW_HASKELL__)
+        else if use_mmap then
+            do hClose h
+               liftM MMappedPackedString (mmapFilePS f)
+#endif
+        else do let read_rest = do
+                        -- We might be making too big a fp here
+                        fp <- mallocForeignPtr blocksize
+                        debugForeignPtr fp $ "gzReadFileLazily "++f
+                        lread <- withForeignPtr fp
+                               $ \p -> hGetBuf h p blocksize
+                        case lread of
+                            0 -> return []
+                            l -> do rest <- unsafeInterleaveIO read_rest
+                                    return (PS fp 0 l:rest)
+                rest <- unsafeInterleaveIO read_rest
+                return $ LazyPackedStrings (header:rest)
+    where blocksize = 1024
+
 hGetLittleEndInt :: Handle -> IO Int
 hGetLittleEndInt h = do
     b1 <- ord `liftM` hGetChar h
@@ -741,6 +869,23 @@ hGetLittleEndInt h = do
     b4 <- ord `liftM` hGetChar h
     return $ b1 + 256*b2 + 65536*b3 + 16777216*b4
 
+gzWriteFilePS :: FilePath -> PackedString -> IO ()
+gzWriteFilePS f ps = gzWriteFilePSs f [ps]
+
+gzWriteFilePSs :: FilePath -> [PackedString] -> IO ()
+gzWriteFilePSs f pss  =
+    withCString f $ \fstr -> withCString "wb" $ \wb -> do
+    gzf <- c_gzopen fstr wb
+    when (gzf == nullPtr) $ fail $ "problem gzopening file for write: "++f
+    mapM_ (gzWriteToGzf gzf) pss `catch`
+              \_ -> fail $ "problem gzwriting file: "++f
+    c_gzclose gzf
+
+gzWriteToGzf :: Ptr () -> PackedString -> IO ()
+gzWriteToGzf gzf (PS x s l) = do
+    lw <- withForeignPtr x $ \p -> c_gzwrite gzf (p `plusPtr` s) l
+    when (lw /= l) $ fail $ "problem in gzWriteToGzf"
+
 -- -----------------------------------------------------------------------------
 -- mmapFilePS
 
@@ -761,12 +906,14 @@ mmapFilePS f = if use_mmap
                        return $ PS fp 0 l
                else readFilePS f
 
+#if defined(__GLASGOW_HASKELL__)
 foreign import ccall unsafe "static fpstring.h my_mmap" my_mmap
     :: Int -> Int -> IO (Ptr Word8)
 foreign import ccall unsafe "static sys/mman.h munmap" c_munmap
     :: Ptr Word8 -> Int -> IO Int
 foreign import ccall unsafe "static unistd.h close" c_close
     :: Int -> IO Int
+#endif
 
 mmap :: FilePath -> IO (ForeignPtr Word8, Int)
 mmap f = do
@@ -780,17 +927,26 @@ mmap f = do
                withForeignPtr thefp $ \p-> hGetBuf h p l
                hClose h
                return (thefp, l)
-       else do fd <- fdToInt `liftM` handleToFd h
+       else do
+#if defined(__GLASGOW_HASKELL__)
+               fd <- fromIntegral `liftM` handleToFd h
                p <- my_mmap l fd
                fp <- if p == nullPtr
-                     then do thefp <- mallocForeignPtr l
+                     then
+#else
+               fp <-
+#endif
+                          do thefp <- mallocForeignPtr l
                              debugForeignPtr thefp $ "mmap short file "++f
                              withForeignPtr thefp $ \p' -> hGetBuf h p' l
                              return thefp
-                     else do fp <- newForeignPtr p (do {c_munmap p l; return (); })
+#if defined(__GLASGOW_HASKELL__)
+                     else do
+                             fp <- FC.newForeignPtr p (do {c_munmap p l; return (); })
                              debugForeignPtr fp $ "mmap "++f
                              return fp
                c_close fd
+#endif
                hClose h
                return (fp, l)
     where mmap_limit = 16*1024
@@ -839,18 +995,6 @@ fromHex2PS :: PackedString -> PackedString
 fromHex2PS (PS x s l) = createPS (l `div` 2) $ \p -> withForeignPtr x $ \f ->
            conv_from_hex p (f `plusPtr` s) (l `div` 2)
 
--- -------------------------------------------------------------------------
--- mylexPS
-
--- | mylexPS reads a single token from the PackedString, where a token is
--- defined as a sequence of non-whitespace characters which are delimited
--- by whitespace.
-
-mylexPS :: PackedString -> Maybe (PackedString,PackedString)
-mylexPS s = if nullPS (dropWhitePS s)
-            then Nothing
-            else Just $ breakWhitePS $ dropWhitePS s
-
 -- -------------------------------------------------------------------------
 -- betweenLinesPS
 
@@ -867,3 +1011,44 @@ betweenLinesPS start end ps
                _ -> Nothing
        _ -> Nothing
 
+-- -------------------------------------------------------------------------
+-- break_after_nth_newline
+
+break_after_nth_newline :: Int -> PackedString
+                        -> Maybe (PackedString, PackedString)
+break_after_nth_newline 0 the_ps | nullPS the_ps = Just (nilPS, nilPS)
+break_after_nth_newline n the_ps@(PS fp the_s l)
+ = unsafePerformIO $ withForeignPtr fp $ \p ->
+   do let findit 0 s | s == end = return $ Just (the_ps, nilPS)
+          findit _ s | s == end = return Nothing
+          findit 0 s = let left_l = s - the_s
+                       in return $ Just (PS fp the_s left_l,
+                                         PS fp s (l - left_l))
+          findit i s = do w <- peekElemOff p s
+                          if w == nl then findit (i-1) (s+1)
+                                     else findit i (s+1)
+          nl = c2w '\n'
+          end = the_s + l
+      findit n the_s
+
+-- -------------------------------------------------------------------------
+-- break_before_nth_newline
+
+break_before_nth_newline :: Int -> PackedString -> (PackedString, PackedString)
+break_before_nth_newline 0 the_ps
+ | nullPS the_ps = (nilPS, nilPS)
+break_before_nth_newline n the_ps@(PS fp the_s l)
+ = unsafePerformIO $ withForeignPtr fp $ \p ->
+   do let findit _ s | s == end = return (the_ps, nilPS)
+          findit i s = do w <- peekElemOff p s
+                          if w == nl
+                            then if i == 0
+                                 then let left_l = s - the_s
+                                      in return (PS fp the_s left_l,
+                                                 PS fp s (l - left_l))
+                                 else findit (i-1) (s+1)
+                            else findit i (s+1)
+          nl = c2w '\n'
+          end = the_s + l
+      findit n the_s
+
