@@ -4,6 +4,9 @@ static const char context_string_value[] =
  "\n"
  "Context:\n"
  "\n"
+ "[update web page to reflect new stable release (1.0.3)\n"
+ "Tomasz Zielonka <tomasz.zielonka@gmail.com>**20050524225643] \n"
+ "\n"
  "[TAG 1.0.3\n"
  "Tomasz Zielonka <tomasz.zielonka@gmail.com>**20050524215127] \n"
  "";
