@@ -15,6 +15,7 @@
  *
  */
 #include <stdlib.h>
+#include <string.h>
 #include <stdio.h>
 #include <sys/types.h>
 
@@ -26,44 +27,37 @@
 
 #include "fpstring.h"
 
-static int my_is_space(char c) {
-  /* I don't use the C library isspace function because I don't want to
-     have the answer depend on the locale, or a patch file could mean
-     different things on different computers.
+/* A locale-independent isspace(3) so patches are interpreted the same
+ * everywhere. */
+#define ISSPACE(c) \
+    ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r')
 
-     Also, as long as I'm writing my own isspace, I've decided to leave out
-     '\f' and '\v' for just a tad more speed, and since I don't anticipate
-     using these characters in patches.
+int first_white(const char *s, int len)
+{
+    const char *start;
+    const char *end;
 
-     And yes, I'm assuming the compiler will inline this function.
-  */
-  return c == ' ' || c == '\n' || c == '\t' || c == '\r';
+    for (start = s, end = s + len; s < end && !ISSPACE(*s); s++);
+
+    return s - start;
 }
 
-int first_white(const char *s, int len) {
-  int i;
-  for (i=0;i<len;i++)
-    if (my_is_space(s[i])) return i;
-  return len;
+int first_nonwhite(const char *s, int len)
+{
+    const char *start;
+    const char *end;
+
+    for (start = s, end = s + len; s < end && ISSPACE(*s); s++);
+
+    return s - start;
 }
 
-int first_nonwhite(const char *s, int len) {
-  int i;
-  for (i=0;i<len;i++)
-    if (!my_is_space(s[i])) return i;
-  return len;
-}
-
-static int is_funky(char c) {
-  /* these are chars we don't want in text files. */
-  return c == 0 || c == 26;
-}
-
-int has_funky_char(const char *s, int len) {
-  int i;
-  for (i=0;i<len;i++)
-    if (is_funky(s[i])) return 1;
-  return 0;
+int has_funky_char(const char *s, int len)
+{
+    /* We'll assume memchr(3) is faster than a simplistic loop that
+     * checks each char against the two sought values, and NUL is more
+     * likely than DOS's Ctrl-Z. */
+    return !!(memchr(s, 0, len) || memchr(s, 26, len));
 }
 
 // mmapping...
@@ -99,29 +93,6 @@ char *my_mmap(int len, int fd) {
 
 #endif
 
-#define GAP (500*1024*1024)
-#define MAX(a,b) ((a>b)?a:b)
-static char *border = NULL;
-
-char *dumb_malloc(int len) {
-  char *filler, *out;
-  if (!border) {
-    filler = (char *) malloc(GAP);
-    border = filler + GAP;
-    free(filler);
-  }
-  filler = (char *) malloc(MAX(GAP,len));
-  if (filler >= border) {
-    filler = (char *) realloc(filler,len);
-    if (filler < border) printf("Argh!!!\n");
-    return filler;
-  }
-  out = (char *) malloc(MAX(GAP,len));
-  out = (char *) realloc(out,len);
-  free(filler);
-  return out;
-}
-
 // ForeignPtr debugging stuff...
 
 static int num_alloced = 0;
@@ -219,28 +190,32 @@ int utf8_to_ints(HsInt *pwc, const unsigned char *s, int n) {
 
 /* Conversion to and from hex */
 
-static unsigned char num2hex(unsigned char c) {
-  if (c < 10) return '0' + c;
-  else return ('a' - 10) + c;
+void conv_to_hex(unsigned char *dest, unsigned char *from, int num_chars)
+{
+    static char hex[] = "0123456789abcdef";
+    unsigned char *end;
+
+    for (end = from + num_chars; from < end; from++) {
+        *dest++ = hex[*from >> 4];
+        *dest++ = hex[*from & 0xf];
+    }
+
+    return;
 }
 
-static unsigned char hex2num(unsigned char c) {
-  if (c - '0' >= 0 && c - '0' < 10) return c - '0';
-  else return c - ('a' - 10);
-}
+#define NYBBLE_TO_INT(c) \
+    ((c) - ((c) >= 'a' ? 'a' - 10 : '0'))
 
-void conv_to_hex(unsigned char *dest, unsigned char *from, int num_chars) {
-  int i;
-  for (i=0;i<num_chars;i++) {
-    const unsigned char thec = from[i];
-    dest[i*2] = num2hex(thec/16);
-    dest[i*2+1] = num2hex(thec%16);
-  }
-}
+void conv_from_hex(unsigned char *dest, unsigned char *from, int num_chars)
+{
+    unsigned char *end;
+    unsigned char c;
 
-void conv_from_hex(unsigned char *dest, unsigned char *from, int num_chars) {
-  int i;
-  for (i=0;i<num_chars;i++)
-    dest[i] = hex2num(from[i*2])*16 + hex2num(from[i*2+1]);
-}
+    end = dest + num_chars;
+    while (dest < end) {
+        c = NYBBLE_TO_INT(*from) << 4, from++;
+        *dest++ = c | NYBBLE_TO_INT(*from), from++;
+    }
 
+    return;
+}
