@@ -16,7 +16,8 @@
 
 module Main (main) where
 
-import Repository ( read_repo )
+import DarcsUtils ( formatPath )
+import DarcsRepo ( read_repo )
 import PatchInfo ( pi_author )
 
 import Data.List ( sort, group )
@@ -31,7 +32,8 @@ intro = do
     "number of patches they have contributed.  This list is automatically\n"++
     "created from the darcs repository, so if there are problems with it,\n"++
     "changes should be made to list_authors.hs.\n"++
-    "Say \"" ++ prog_name ++ " stats\" for more detailed information.\n"++
+    "\n"++
+    "Run " ++ formatPath (prog_name ++ " stats") ++ " for more detailed information.\n"++
     "\n"
 
 main :: IO ()
@@ -42,10 +44,11 @@ main = do darcs_history <- read_repo "."
                          $ concat darcs_history
 
 droundy :: String
-droundy = "David Roundy <droundy@abridgegame.org>"
+droundy = "David Roundy <droundy@darcs.net>"
 
 canonize_author :: String -> String
 canonize_author "David" = droundy
+canonize_author "jani@iv.ro" = "Jani Monoses"
 canonize_author "brian@brianweb.net" = "Brian Alliet <brian@brianweb.net>"
 canonize_author "jan@informatik.unibw-muenchen.de"
     = "Jan Scheffczyk <jan@informatik.unibw-muenchen.de>"
@@ -71,6 +74,7 @@ canonize_author "wnoise@ofb.net" = "Aaron Denney <wnoise@ofb.net>"
 canonize_author "kannan@cakoose.com" = "Kannan Goundan <kannan@cakoose.com>"
 canonize_author "zander@kde.org" = "Thomas Zander <zander@kde.org>"
 canonize_author "rho@swiftdsl.com.au" = "Nigel Rowe <rho@swiftdsl.com.au>"
+canonize_author "simons@cryp.to" = "Peter Simons <simons@cryp.to>"
 canonize_author "ptp@lysator.liu.se" = "Tommy Pettersson <ptp@lysator.liu.se>"
 canonize_author s
     | s `contains` "roundy" = droundy
