@@ -11,7 +11,7 @@ use Shell::Command;
 
 use strict;
 
-rm_rf  'temp1';
+cleanup  'temp1';
 mkpath 'temp1';
 chdir  'temp1';
 darcs 'init';
@@ -56,8 +56,25 @@ $test_name = 'Make sure that darcs doesn\'t complains about duplicate adds when
 mkpath 'c.d/';
 touch 'c.d/baz';
 like(darcs(qw( add c.d/baz c.d )), qr/^$/, $test_name);
+
 ###
 
+$test_name = 'Make sure that add output looks good when adding files in subdir.';
+mkpath 'd.d/';
+touch 'd.d/foo';
+like(darcs(qw(add -rv d.d)), qr/d.d\/foo/,$test_name);
+
+TODO: {
+    local $TODO = 'waiting on coding';
+    my $test_name = "add should fail on files it can't read (because it would fail to record it later anyway).";
+    touch "no_perms.txt";
+    chmod(0000,"no_perms.txt");
+    like(darcs(qw(add no_perms.txt)), qr/permission denied/,$test_name);
+    rm_rf "no_perms.txt";
+
+}
+
+
 {
     my $test_name = 'adding a non-existent dir and file gives the expected message';
     my $out = darcs(qw!add notadir/notafile!);
@@ -66,5 +83,5 @@ like(darcs(qw( add c.d/baz c.d )), qr/^$/, $test_name);
 
 
 chdir '../';
-rm_rf 'temp1';
+cleanup 'temp1';
 ok((!-d 'temp1'), 'temp1 directory was deleted');
