@@ -1,12 +1,12 @@
 #!/bin/sh
 set -ev
 
-DARCS=$PWD/../darcs
+test $DARCS || DARCS=$PWD/../darcs
 
 rm -rf temp1
 mkdir temp1
 cd temp1
-$DARCS inittree
+$DARCS init
 touch foo
 $DARCS add foo
 $DARCS record -a -m add_foo -A x
@@ -14,3 +14,4 @@ rm foo
 $DARCS whatsnew
 cd ..
 rm -rf temp1
+
