@@ -11,7 +11,7 @@ use Shell::Command;
 
 use strict;
 
-rm_rf  'temp1';
+cleanup  'temp1';
 mkpath 'temp1';
 chdir  'temp1';
 darcs 'init';
@@ -31,5 +31,5 @@ TODO: {
 ####
 
 chdir '../';
-rm_rf 'temp1';
+cleanup 'temp1';
 ok((!-d 'temp1'), 'temp1 directory was deleted');
